from __future__ import annotations

from contextlib import suppress
from typing import Any

from beartype.door import die_if_unbearable
from beartype.roar import BeartypeDoorHintViolation
from hypothesis import Phase
from hypothesis import example
from hypothesis import given
from hypothesis import settings
from numpy import array
from numpy import empty
from numpy import nan
from numpy import zeros
from pytest import mark
from pytest import param

from utilities.hypothesis.numpy import float_arrays
from utilities.hypothesis.numpy import int_arrays
from utilities.numpy.typing import NDArray0
from utilities.numpy.typing import NDArray1
from utilities.numpy.typing import NDArray2
from utilities.numpy.typing import NDArray3
from utilities.numpy.typing import NDArrayB
from utilities.numpy.typing import NDArrayB0
from utilities.numpy.typing import NDArrayB1
from utilities.numpy.typing import NDArrayB2
from utilities.numpy.typing import NDArrayB3
from utilities.numpy.typing import NDArrayD0
from utilities.numpy.typing import NDArrayD1
from utilities.numpy.typing import NDArrayD2
from utilities.numpy.typing import NDArrayD3
from utilities.numpy.typing import NDArrayDas
from utilities.numpy.typing import NDArrayDas0
from utilities.numpy.typing import NDArrayDas1
from utilities.numpy.typing import NDArrayDas2
from utilities.numpy.typing import NDArrayDas3
from utilities.numpy.typing import NDArrayDD
from utilities.numpy.typing import NDArrayDD0
from utilities.numpy.typing import NDArrayDD1
from utilities.numpy.typing import NDArrayDD2
from utilities.numpy.typing import NDArrayDD3
from utilities.numpy.typing import NDArrayDfs
from utilities.numpy.typing import NDArrayDfs0
from utilities.numpy.typing import NDArrayDfs1
from utilities.numpy.typing import NDArrayDfs2
from utilities.numpy.typing import NDArrayDfs3
from utilities.numpy.typing import NDArrayDh
from utilities.numpy.typing import NDArrayDh0
from utilities.numpy.typing import NDArrayDh1
from utilities.numpy.typing import NDArrayDh2
from utilities.numpy.typing import NDArrayDh3
from utilities.numpy.typing import NDArrayDM
from utilities.numpy.typing import NDArrayDm
from utilities.numpy.typing import NDArrayDM0
from utilities.numpy.typing import NDArrayDm0
from utilities.numpy.typing import NDArrayDM1
from utilities.numpy.typing import NDArrayDm1
from utilities.numpy.typing import NDArrayDM2
from utilities.numpy.typing import NDArrayDm2
from utilities.numpy.typing import NDArrayDM3
from utilities.numpy.typing import NDArrayDm3
from utilities.numpy.typing import NDArrayDms
from utilities.numpy.typing import NDArrayDms0
from utilities.numpy.typing import NDArrayDms1
from utilities.numpy.typing import NDArrayDms2
from utilities.numpy.typing import NDArrayDms3
from utilities.numpy.typing import NDArrayDns
from utilities.numpy.typing import NDArrayDns0
from utilities.numpy.typing import NDArrayDns1
from utilities.numpy.typing import NDArrayDns2
from utilities.numpy.typing import NDArrayDns3
from utilities.numpy.typing import NDArrayDps
from utilities.numpy.typing import NDArrayDps0
from utilities.numpy.typing import NDArrayDps1
from utilities.numpy.typing import NDArrayDps2
from utilities.numpy.typing import NDArrayDps3
from utilities.numpy.typing import NDArrayDs
from utilities.numpy.typing import NDArrayDs0
from utilities.numpy.typing import NDArrayDs1
from utilities.numpy.typing import NDArrayDs2
from utilities.numpy.typing import NDArrayDs3
from utilities.numpy.typing import NDArrayDus
from utilities.numpy.typing import NDArrayDus0
from utilities.numpy.typing import NDArrayDus1
from utilities.numpy.typing import NDArrayDus2
from utilities.numpy.typing import NDArrayDus3
from utilities.numpy.typing import NDArrayDW
from utilities.numpy.typing import NDArrayDW0
from utilities.numpy.typing import NDArrayDW1
from utilities.numpy.typing import NDArrayDW2
from utilities.numpy.typing import NDArrayDW3
from utilities.numpy.typing import NDArrayDY
from utilities.numpy.typing import NDArrayDY0
from utilities.numpy.typing import NDArrayDY1
from utilities.numpy.typing import NDArrayDY2
from utilities.numpy.typing import NDArrayDY3
from utilities.numpy.typing import NDArrayF
from utilities.numpy.typing import NDArrayF0
from utilities.numpy.typing import NDArrayF0Fin
from utilities.numpy.typing import NDArrayF0FinInt
from utilities.numpy.typing import NDArrayF0FinIntNan
from utilities.numpy.typing import NDArrayF0FinNan
from utilities.numpy.typing import NDArrayF0FinNeg
from utilities.numpy.typing import NDArrayF0FinNegNan
from utilities.numpy.typing import NDArrayF0FinNonNeg
from utilities.numpy.typing import NDArrayF0FinNonNegNan
from utilities.numpy.typing import NDArrayF0FinNonPos
from utilities.numpy.typing import NDArrayF0FinNonPosNan
from utilities.numpy.typing import NDArrayF0FinNonZr
from utilities.numpy.typing import NDArrayF0FinNonZrNan
from utilities.numpy.typing import NDArrayF0FinPos
from utilities.numpy.typing import NDArrayF0FinPosNan
from utilities.numpy.typing import NDArrayF0Int
from utilities.numpy.typing import NDArrayF0IntNan
from utilities.numpy.typing import NDArrayF0Neg
from utilities.numpy.typing import NDArrayF0NegNan
from utilities.numpy.typing import NDArrayF0NonNeg
from utilities.numpy.typing import NDArrayF0NonNegNan
from utilities.numpy.typing import NDArrayF0NonPos
from utilities.numpy.typing import NDArrayF0NonPosNan
from utilities.numpy.typing import NDArrayF0NonZr
from utilities.numpy.typing import NDArrayF0NonZrNan
from utilities.numpy.typing import NDArrayF0Pos
from utilities.numpy.typing import NDArrayF0PosNan
from utilities.numpy.typing import NDArrayF0Zr
from utilities.numpy.typing import NDArrayF0ZrFinNonMic
from utilities.numpy.typing import NDArrayF0ZrFinNonMicNan
from utilities.numpy.typing import NDArrayF0ZrNan
from utilities.numpy.typing import NDArrayF0ZrNonMic
from utilities.numpy.typing import NDArrayF0ZrNonMicNan
from utilities.numpy.typing import NDArrayF1
from utilities.numpy.typing import NDArrayF1Fin
from utilities.numpy.typing import NDArrayF1FinInt
from utilities.numpy.typing import NDArrayF1FinIntNan
from utilities.numpy.typing import NDArrayF1FinNan
from utilities.numpy.typing import NDArrayF1FinNeg
from utilities.numpy.typing import NDArrayF1FinNegNan
from utilities.numpy.typing import NDArrayF1FinNonNeg
from utilities.numpy.typing import NDArrayF1FinNonNegNan
from utilities.numpy.typing import NDArrayF1FinNonPos
from utilities.numpy.typing import NDArrayF1FinNonPosNan
from utilities.numpy.typing import NDArrayF1FinNonZr
from utilities.numpy.typing import NDArrayF1FinNonZrNan
from utilities.numpy.typing import NDArrayF1FinPos
from utilities.numpy.typing import NDArrayF1FinPosNan
from utilities.numpy.typing import NDArrayF1Int
from utilities.numpy.typing import NDArrayF1IntNan
from utilities.numpy.typing import NDArrayF1Neg
from utilities.numpy.typing import NDArrayF1NegNan
from utilities.numpy.typing import NDArrayF1NonNeg
from utilities.numpy.typing import NDArrayF1NonNegNan
from utilities.numpy.typing import NDArrayF1NonPos
from utilities.numpy.typing import NDArrayF1NonPosNan
from utilities.numpy.typing import NDArrayF1NonZr
from utilities.numpy.typing import NDArrayF1NonZrNan
from utilities.numpy.typing import NDArrayF1Pos
from utilities.numpy.typing import NDArrayF1PosNan
from utilities.numpy.typing import NDArrayF1Zr
from utilities.numpy.typing import NDArrayF1ZrFinNonMic
from utilities.numpy.typing import NDArrayF1ZrFinNonMicNan
from utilities.numpy.typing import NDArrayF1ZrNan
from utilities.numpy.typing import NDArrayF1ZrNonMic
from utilities.numpy.typing import NDArrayF1ZrNonMicNan
from utilities.numpy.typing import NDArrayF2
from utilities.numpy.typing import NDArrayF2Fin
from utilities.numpy.typing import NDArrayF2FinInt
from utilities.numpy.typing import NDArrayF2FinIntNan
from utilities.numpy.typing import NDArrayF2FinNan
from utilities.numpy.typing import NDArrayF2FinNeg
from utilities.numpy.typing import NDArrayF2FinNegNan
from utilities.numpy.typing import NDArrayF2FinNonNeg
from utilities.numpy.typing import NDArrayF2FinNonNegNan
from utilities.numpy.typing import NDArrayF2FinNonPos
from utilities.numpy.typing import NDArrayF2FinNonPosNan
from utilities.numpy.typing import NDArrayF2FinNonZr
from utilities.numpy.typing import NDArrayF2FinNonZrNan
from utilities.numpy.typing import NDArrayF2FinPos
from utilities.numpy.typing import NDArrayF2FinPosNan
from utilities.numpy.typing import NDArrayF2Int
from utilities.numpy.typing import NDArrayF2IntNan
from utilities.numpy.typing import NDArrayF2Neg
from utilities.numpy.typing import NDArrayF2NegNan
from utilities.numpy.typing import NDArrayF2NonNeg
from utilities.numpy.typing import NDArrayF2NonNegNan
from utilities.numpy.typing import NDArrayF2NonPos
from utilities.numpy.typing import NDArrayF2NonPosNan
from utilities.numpy.typing import NDArrayF2NonZr
from utilities.numpy.typing import NDArrayF2NonZrNan
from utilities.numpy.typing import NDArrayF2Pos
from utilities.numpy.typing import NDArrayF2PosNan
from utilities.numpy.typing import NDArrayF2Zr
from utilities.numpy.typing import NDArrayF2ZrFinNonMic
from utilities.numpy.typing import NDArrayF2ZrFinNonMicNan
from utilities.numpy.typing import NDArrayF2ZrNan
from utilities.numpy.typing import NDArrayF2ZrNonMic
from utilities.numpy.typing import NDArrayF2ZrNonMicNan
from utilities.numpy.typing import NDArrayF3
from utilities.numpy.typing import NDArrayF3Fin
from utilities.numpy.typing import NDArrayF3FinInt
from utilities.numpy.typing import NDArrayF3FinIntNan
from utilities.numpy.typing import NDArrayF3FinNan
from utilities.numpy.typing import NDArrayF3FinNeg
from utilities.numpy.typing import NDArrayF3FinNegNan
from utilities.numpy.typing import NDArrayF3FinNonNeg
from utilities.numpy.typing import NDArrayF3FinNonNegNan
from utilities.numpy.typing import NDArrayF3FinNonPos
from utilities.numpy.typing import NDArrayF3FinNonPosNan
from utilities.numpy.typing import NDArrayF3FinNonZr
from utilities.numpy.typing import NDArrayF3FinNonZrNan
from utilities.numpy.typing import NDArrayF3FinPos
from utilities.numpy.typing import NDArrayF3FinPosNan
from utilities.numpy.typing import NDArrayF3Int
from utilities.numpy.typing import NDArrayF3IntNan
from utilities.numpy.typing import NDArrayF3Neg
from utilities.numpy.typing import NDArrayF3NegNan
from utilities.numpy.typing import NDArrayF3NonNeg
from utilities.numpy.typing import NDArrayF3NonNegNan
from utilities.numpy.typing import NDArrayF3NonPos
from utilities.numpy.typing import NDArrayF3NonPosNan
from utilities.numpy.typing import NDArrayF3NonZr
from utilities.numpy.typing import NDArrayF3NonZrNan
from utilities.numpy.typing import NDArrayF3Pos
from utilities.numpy.typing import NDArrayF3PosNan
from utilities.numpy.typing import NDArrayF3Zr
from utilities.numpy.typing import NDArrayF3ZrFinNonMic
from utilities.numpy.typing import NDArrayF3ZrFinNonMicNan
from utilities.numpy.typing import NDArrayF3ZrNan
from utilities.numpy.typing import NDArrayF3ZrNonMic
from utilities.numpy.typing import NDArrayF3ZrNonMicNan
from utilities.numpy.typing import NDArrayFFin
from utilities.numpy.typing import NDArrayFFinInt
from utilities.numpy.typing import NDArrayFFinIntNan
from utilities.numpy.typing import NDArrayFFinNan
from utilities.numpy.typing import NDArrayFFinNeg
from utilities.numpy.typing import NDArrayFFinNegNan
from utilities.numpy.typing import NDArrayFFinNonNeg
from utilities.numpy.typing import NDArrayFFinNonNegNan
from utilities.numpy.typing import NDArrayFFinNonPos
from utilities.numpy.typing import NDArrayFFinNonPosNan
from utilities.numpy.typing import NDArrayFFinNonZr
from utilities.numpy.typing import NDArrayFFinNonZrNan
from utilities.numpy.typing import NDArrayFFinPos
from utilities.numpy.typing import NDArrayFFinPosNan
from utilities.numpy.typing import NDArrayFInt
from utilities.numpy.typing import NDArrayFIntNan
from utilities.numpy.typing import NDArrayFNeg
from utilities.numpy.typing import NDArrayFNegNan
from utilities.numpy.typing import NDArrayFNonNeg
from utilities.numpy.typing import NDArrayFNonNegNan
from utilities.numpy.typing import NDArrayFNonPos
from utilities.numpy.typing import NDArrayFNonPosNan
from utilities.numpy.typing import NDArrayFNonZr
from utilities.numpy.typing import NDArrayFNonZrNan
from utilities.numpy.typing import NDArrayFPos
from utilities.numpy.typing import NDArrayFPosNan
from utilities.numpy.typing import NDArrayFZr
from utilities.numpy.typing import NDArrayFZrFinNonMic
from utilities.numpy.typing import NDArrayFZrFinNonMicNan
from utilities.numpy.typing import NDArrayFZrNan
from utilities.numpy.typing import NDArrayFZrNonMic
from utilities.numpy.typing import NDArrayFZrNonMicNan
from utilities.numpy.typing import NDArrayI
from utilities.numpy.typing import NDArrayI0
from utilities.numpy.typing import NDArrayI0Neg
from utilities.numpy.typing import NDArrayI0NonNeg
from utilities.numpy.typing import NDArrayI0NonPos
from utilities.numpy.typing import NDArrayI0NonZr
from utilities.numpy.typing import NDArrayI0Pos
from utilities.numpy.typing import NDArrayI0Zr
from utilities.numpy.typing import NDArrayI1
from utilities.numpy.typing import NDArrayI1Neg
from utilities.numpy.typing import NDArrayI1NonNeg
from utilities.numpy.typing import NDArrayI1NonPos
from utilities.numpy.typing import NDArrayI1NonZr
from utilities.numpy.typing import NDArrayI1Pos
from utilities.numpy.typing import NDArrayI1Zr
from utilities.numpy.typing import NDArrayI2
from utilities.numpy.typing import NDArrayI2Neg
from utilities.numpy.typing import NDArrayI2NonNeg
from utilities.numpy.typing import NDArrayI2NonPos
from utilities.numpy.typing import NDArrayI2NonZr
from utilities.numpy.typing import NDArrayI2Pos
from utilities.numpy.typing import NDArrayI2Zr
from utilities.numpy.typing import NDArrayI3
from utilities.numpy.typing import NDArrayI3Neg
from utilities.numpy.typing import NDArrayI3NonNeg
from utilities.numpy.typing import NDArrayI3NonPos
from utilities.numpy.typing import NDArrayI3NonZr
from utilities.numpy.typing import NDArrayI3Pos
from utilities.numpy.typing import NDArrayI3Zr
from utilities.numpy.typing import NDArrayINeg
from utilities.numpy.typing import NDArrayINonNeg
from utilities.numpy.typing import NDArrayINonPos
from utilities.numpy.typing import NDArrayINonZr
from utilities.numpy.typing import NDArrayIPos
from utilities.numpy.typing import NDArrayIZr
from utilities.numpy.typing import NDArrayO
from utilities.numpy.typing import NDArrayO0
from utilities.numpy.typing import NDArrayO1
from utilities.numpy.typing import NDArrayO2
from utilities.numpy.typing import NDArrayO3
from utilities.numpy.typing import datetime64as
from utilities.numpy.typing import datetime64D
from utilities.numpy.typing import datetime64fs
from utilities.numpy.typing import datetime64h
from utilities.numpy.typing import datetime64M
from utilities.numpy.typing import datetime64m
from utilities.numpy.typing import datetime64ms
from utilities.numpy.typing import datetime64ns
from utilities.numpy.typing import datetime64ps
from utilities.numpy.typing import datetime64s
from utilities.numpy.typing import datetime64us
from utilities.numpy.typing import datetime64W
from utilities.numpy.typing import datetime64Y


class TestHints:
    @mark.parametrize(
        ("dtype", "hint"),
        [
            param(bool, NDArrayB),
            param(datetime64Y, NDArrayDY),
            param(datetime64M, NDArrayDM),
            param(datetime64W, NDArrayDW),
            param(datetime64D, NDArrayDD),
            param(datetime64h, NDArrayDh),
            param(datetime64m, NDArrayDm),
            param(datetime64s, NDArrayDs),
            param(datetime64ms, NDArrayDms),
            param(datetime64us, NDArrayDus),
            param(datetime64ns, NDArrayDns),
            param(datetime64ps, NDArrayDps),
            param(datetime64fs, NDArrayDfs),
            param(datetime64as, NDArrayDas),
            param(float, NDArrayF),
            param(int, NDArrayI),
            param(object, NDArrayO),
        ],
    )
    def test_dtype(self, dtype: Any, hint: Any) -> None:
        arr = empty(0, dtype=dtype)
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("ndim", "hint"),
        [
            param(0, NDArray0),
            param(1, NDArray1),
            param(2, NDArray2),
            param(3, NDArray3),
        ],
    )
    def test_ndim(self, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=float)
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("dtype", "ndim", "hint"),
        [
            # ndim 0
            param(bool, 0, NDArrayB0),
            param(datetime64D, 0, NDArrayD0),
            param(datetime64Y, 0, NDArrayDY0),
            param(datetime64M, 0, NDArrayDM0),
            param(datetime64W, 0, NDArrayDW0),
            param(datetime64D, 0, NDArrayDD0),
            param(datetime64h, 0, NDArrayDh0),
            param(datetime64m, 0, NDArrayDm0),
            param(datetime64s, 0, NDArrayDs0),
            param(datetime64ms, 0, NDArrayDms0),
            param(datetime64us, 0, NDArrayDus0),
            param(datetime64ns, 0, NDArrayDns0),
            param(datetime64ps, 0, NDArrayDps0),
            param(datetime64fs, 0, NDArrayDfs0),
            param(datetime64as, 0, NDArrayDas0),
            param(float, 0, NDArrayF0),
            param(int, 0, NDArrayI0),
            param(object, 0, NDArrayO0),
            # ndim 1
            param(bool, 1, NDArrayB1),
            param(datetime64D, 1, NDArrayD1),
            param(datetime64Y, 1, NDArrayDY1),
            param(datetime64M, 1, NDArrayDM1),
            param(datetime64W, 1, NDArrayDW1),
            param(datetime64D, 1, NDArrayDD1),
            param(datetime64h, 1, NDArrayDh1),
            param(datetime64m, 1, NDArrayDm1),
            param(datetime64s, 1, NDArrayDs1),
            param(datetime64ms, 1, NDArrayDms1),
            param(datetime64us, 1, NDArrayDus1),
            param(datetime64ns, 1, NDArrayDns1),
            param(datetime64ps, 1, NDArrayDps1),
            param(datetime64fs, 1, NDArrayDfs1),
            param(datetime64as, 1, NDArrayDas1),
            param(float, 1, NDArrayF1),
            param(int, 1, NDArrayI1),
            param(object, 1, NDArrayO1),
            # ndim 2
            param(bool, 2, NDArrayB2),
            param(datetime64D, 2, NDArrayD2),
            param(datetime64Y, 2, NDArrayDY2),
            param(datetime64M, 2, NDArrayDM2),
            param(datetime64W, 2, NDArrayDW2),
            param(datetime64D, 2, NDArrayDD2),
            param(datetime64h, 2, NDArrayDh2),
            param(datetime64m, 2, NDArrayDm2),
            param(datetime64s, 2, NDArrayDs2),
            param(datetime64ms, 2, NDArrayDms2),
            param(datetime64us, 2, NDArrayDus2),
            param(datetime64ns, 2, NDArrayDns2),
            param(datetime64ps, 2, NDArrayDps2),
            param(datetime64fs, 2, NDArrayDfs2),
            param(datetime64as, 2, NDArrayDas2),
            param(float, 2, NDArrayF2),
            param(int, 2, NDArrayI2),
            param(object, 2, NDArrayO2),
            # ndim 3
            param(bool, 3, NDArrayB3),
            param(datetime64D, 3, NDArrayD3),
            param(datetime64Y, 3, NDArrayDY3),
            param(datetime64M, 3, NDArrayDM3),
            param(datetime64W, 3, NDArrayDW3),
            param(datetime64D, 3, NDArrayDD3),
            param(datetime64h, 3, NDArrayDh3),
            param(datetime64m, 3, NDArrayDm3),
            param(datetime64s, 3, NDArrayDs3),
            param(datetime64ms, 3, NDArrayDms3),
            param(datetime64us, 3, NDArrayDus3),
            param(datetime64ns, 3, NDArrayDns3),
            param(datetime64ps, 3, NDArrayDps3),
            param(datetime64fs, 3, NDArrayDfs3),
            param(datetime64as, 3, NDArrayDas3),
            param(float, 3, NDArrayF3),
            param(int, 3, NDArrayI3),
            param(object, 3, NDArrayO3),
        ],
    )
    def test_compound(self, dtype: Any, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=dtype)
        die_if_unbearable(arr, hint)

    @given(arr=int_arrays())
    @example(arr=array([], dtype=int))
    @mark.parametrize("dtype", [param(int), param(float)])
    @mark.parametrize(
        "hint",
        [
            param(NDArrayINeg),
            param(NDArrayINonNeg),
            param(NDArrayINonPos),
            param(NDArrayINonZr),
            param(NDArrayIPos),
            param(NDArrayIZr),
            param(NDArrayI0Neg),
            param(NDArrayI0NonNeg),
            param(NDArrayI0NonPos),
            param(NDArrayI0NonZr),
            param(NDArrayI0Pos),
            param(NDArrayI0Zr),
            param(NDArrayI1Neg),
            param(NDArrayI1NonNeg),
            param(NDArrayI1NonPos),
            param(NDArrayI1NonZr),
            param(NDArrayI1Pos),
            param(NDArrayI1Zr),
            param(NDArrayI2Neg),
            param(NDArrayI2NonNeg),
            param(NDArrayI2NonPos),
            param(NDArrayI2NonZr),
            param(NDArrayI2Pos),
            param(NDArrayI2Zr),
            param(NDArrayI3Neg),
            param(NDArrayI3NonNeg),
            param(NDArrayI3NonPos),
            param(NDArrayI3NonZr),
            param(NDArrayI3Pos),
            param(NDArrayI3Zr),
        ],
    )
    @settings(max_examples=1, phases={Phase.explicit, Phase.generate})
    def test_int_checks(self, arr: NDArrayI, dtype: Any, hint: Any) -> None:
        with suppress(BeartypeDoorHintViolation):
            die_if_unbearable(arr.astype(dtype), hint)

    @given(arr=float_arrays())
    @example(arr=array([], dtype=float))
    @example(arr=array([nan], dtype=float))
    @example(arr=array([nan, nan], dtype=float))
    @mark.parametrize(
        "hint",
        [
            param(NDArrayFFin),
            param(NDArrayFFinInt),
            param(NDArrayFFinIntNan),
            param(NDArrayFFinNeg),
            param(NDArrayFFinNegNan),
            param(NDArrayFFinNonNeg),
            param(NDArrayFFinNonNegNan),
            param(NDArrayFFinNonPos),
            param(NDArrayFFinNonPosNan),
            param(NDArrayFFinNonZr),
            param(NDArrayFFinNonZrNan),
            param(NDArrayFFinPos),
            param(NDArrayFFinPosNan),
            param(NDArrayFFinNan),
            param(NDArrayFInt),
            param(NDArrayFIntNan),
            param(NDArrayFNeg),
            param(NDArrayFNegNan),
            param(NDArrayFNonNeg),
            param(NDArrayFNonNegNan),
            param(NDArrayFNonPos),
            param(NDArrayFNonPosNan),
            param(NDArrayFNonZr),
            param(NDArrayFNonZrNan),
            param(NDArrayFPos),
            param(NDArrayFPosNan),
            param(NDArrayFZr),
            param(NDArrayFZrNonMic),
            param(NDArrayFZrNonMicNan),
            param(NDArrayFZrNan),
            param(NDArrayFZrFinNonMic),
            param(NDArrayFZrFinNonMicNan),
            param(NDArrayF0Fin),
            param(NDArrayF0FinInt),
            param(NDArrayF0FinIntNan),
            param(NDArrayF0FinNeg),
            param(NDArrayF0FinNegNan),
            param(NDArrayF0FinNonNeg),
            param(NDArrayF0FinNonNegNan),
            param(NDArrayF0FinNonPos),
            param(NDArrayF0FinNonPosNan),
            param(NDArrayF0FinNonZr),
            param(NDArrayF0FinNonZrNan),
            param(NDArrayF0FinPos),
            param(NDArrayF0FinPosNan),
            param(NDArrayF0FinNan),
            param(NDArrayF0Int),
            param(NDArrayF0IntNan),
            param(NDArrayF0Neg),
            param(NDArrayF0NegNan),
            param(NDArrayF0NonNeg),
            param(NDArrayF0NonNegNan),
            param(NDArrayF0NonPos),
            param(NDArrayF0NonPosNan),
            param(NDArrayF0NonZr),
            param(NDArrayF0NonZrNan),
            param(NDArrayF0Pos),
            param(NDArrayF0PosNan),
            param(NDArrayF0Zr),
            param(NDArrayF0ZrNonMic),
            param(NDArrayF0ZrNonMicNan),
            param(NDArrayF0ZrNan),
            param(NDArrayF0ZrFinNonMic),
            param(NDArrayF0ZrFinNonMicNan),
            param(NDArrayF1Fin),
            param(NDArrayF1FinInt),
            param(NDArrayF1FinIntNan),
            param(NDArrayF1FinNeg),
            param(NDArrayF1FinNegNan),
            param(NDArrayF1FinNonNeg),
            param(NDArrayF1FinNonNegNan),
            param(NDArrayF1FinNonPos),
            param(NDArrayF1FinNonPosNan),
            param(NDArrayF1FinNonZr),
            param(NDArrayF1FinNonZrNan),
            param(NDArrayF1FinPos),
            param(NDArrayF1FinPosNan),
            param(NDArrayF1FinNan),
            param(NDArrayF1Int),
            param(NDArrayF1IntNan),
            param(NDArrayF1Neg),
            param(NDArrayF1NegNan),
            param(NDArrayF1NonNeg),
            param(NDArrayF1NonNegNan),
            param(NDArrayF1NonPos),
            param(NDArrayF1NonPosNan),
            param(NDArrayF1NonZr),
            param(NDArrayF1NonZrNan),
            param(NDArrayF1Pos),
            param(NDArrayF1PosNan),
            param(NDArrayF1Zr),
            param(NDArrayF1ZrNonMic),
            param(NDArrayF1ZrNonMicNan),
            param(NDArrayF1ZrNan),
            param(NDArrayF1ZrFinNonMic),
            param(NDArrayF1ZrFinNonMicNan),
            param(NDArrayF2Fin),
            param(NDArrayF2FinInt),
            param(NDArrayF2FinIntNan),
            param(NDArrayF2FinNeg),
            param(NDArrayF2FinNegNan),
            param(NDArrayF2FinNonNeg),
            param(NDArrayF2FinNonNegNan),
            param(NDArrayF2FinNonPos),
            param(NDArrayF2FinNonPosNan),
            param(NDArrayF2FinNonZr),
            param(NDArrayF2FinNonZrNan),
            param(NDArrayF2FinPos),
            param(NDArrayF2FinPosNan),
            param(NDArrayF2FinNan),
            param(NDArrayF2Int),
            param(NDArrayF2IntNan),
            param(NDArrayF2Neg),
            param(NDArrayF2NegNan),
            param(NDArrayF2NonNeg),
            param(NDArrayF2NonNegNan),
            param(NDArrayF2NonPos),
            param(NDArrayF2NonPosNan),
            param(NDArrayF2NonZr),
            param(NDArrayF2NonZrNan),
            param(NDArrayF2Pos),
            param(NDArrayF2PosNan),
            param(NDArrayF2Zr),
            param(NDArrayF2ZrNonMic),
            param(NDArrayF2ZrNonMicNan),
            param(NDArrayF2ZrNan),
            param(NDArrayF2ZrFinNonMic),
            param(NDArrayF2ZrFinNonMicNan),
            param(NDArrayF3Fin),
            param(NDArrayF3FinInt),
            param(NDArrayF3FinIntNan),
            param(NDArrayF3FinNeg),
            param(NDArrayF3FinNegNan),
            param(NDArrayF3FinNonNeg),
            param(NDArrayF3FinNonNegNan),
            param(NDArrayF3FinNonPos),
            param(NDArrayF3FinNonPosNan),
            param(NDArrayF3FinNonZr),
            param(NDArrayF3FinNonZrNan),
            param(NDArrayF3FinPos),
            param(NDArrayF3FinPosNan),
            param(NDArrayF3FinNan),
            param(NDArrayF3Int),
            param(NDArrayF3IntNan),
            param(NDArrayF3Neg),
            param(NDArrayF3NegNan),
            param(NDArrayF3NonNeg),
            param(NDArrayF3NonNegNan),
            param(NDArrayF3NonPos),
            param(NDArrayF3NonPosNan),
            param(NDArrayF3NonZr),
            param(NDArrayF3NonZrNan),
            param(NDArrayF3Pos),
            param(NDArrayF3PosNan),
            param(NDArrayF3Zr),
            param(NDArrayF3ZrNonMic),
            param(NDArrayF3ZrNonMicNan),
            param(NDArrayF3ZrNan),
            param(NDArrayF3ZrFinNonMic),
            param(NDArrayF3ZrFinNonMicNan),
        ],
    )
    @settings(max_examples=1, phases={Phase.explicit, Phase.generate})
    def test_float_checks(self, arr: NDArrayF, hint: Any) -> None:
        with suppress(BeartypeDoorHintViolation):
            die_if_unbearable(arr, hint)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import dictionaries
from hypothesis.strategies import integers
from hypothesis.strategies import sampled_from
from pytest import mark
from xarray import DataArray

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import text_ascii
from utilities.hypothesis.pandas import int_indexes
from utilities.hypothesis.xarray import float_data_arrays
from utilities.xarray import ewma
from utilities.xarray import exp_moving_sum

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas.typing import IndexA


class TestBottleNeckInstalled:
    def test_main(self) -> None:
        array = DataArray([], {"dim": []}, ["dim"])
        _ = array.ffill(dim="dim")


class TestEwma:
    @given(
        data=data(),
        indexes=dictionaries(
            text_ascii(), int_indexes(), min_size=1, max_size=3
        ),
        halflife=integers(1, 10),
    )
    def test_main(
        self, data: DataObject, indexes: Mapping[str, IndexA], halflife: int
    ) -> None:
        array = data.draw(float_data_arrays(indexes))
        dim = data.draw(sampled_from(list(indexes)))
        with assume_does_not_raise(RuntimeWarning):
            _ = ewma(array, {dim: halflife})


class TestExpMovingSum:
    @given(
        data=data(),
        indexes=dictionaries(
            text_ascii(), int_indexes(), min_size=1, max_size=3
        ),
        halflife=integers(1, 10),
    )
    def test_main(
        self, data: DataObject, indexes: Mapping[str, IndexA], halflife: int
    ) -> None:
        array = data.draw(float_data_arrays(indexes))
        dim = data.draw(sampled_from(list(indexes)))
        with assume_does_not_raise(RuntimeWarning):
            _ = exp_moving_sum(array, {dim: halflife})


class TestNumbaggInstalled:
    @mark.xfail(
        reason="RuntimeError: Cannot install on Python version 3.11.4; "
        "only versions >=3.7,<3.11 are supported."
    )
    def test_main(self) -> None:
        array = DataArray([], {"dim": []}, ["dim"])
        _ = array.rolling_exp(dim=1.0).sum()



from __future__ import annotations

from collections.abc import Callable
from typing import Annotated
from typing import Any
from typing import cast

from beartype.vale import Is
from beartype.vale import IsAttr
from beartype.vale import IsEqual
from numpy import bool_
from numpy import float64
from numpy import int64
from numpy import isfinite
from numpy import log
from numpy import object_
from numpy import unravel_index
from numpy.random import default_rng
from numpy.typing import NDArray

from utilities.beartype import NDim0
from utilities.beartype import NDim1
from utilities.beartype import NDim2
from utilities.beartype import NDim3
from utilities.numpy.checks import is_finite_and_integral
from utilities.numpy.checks import is_finite_and_integral_or_nan
from utilities.numpy.checks import is_finite_and_negative
from utilities.numpy.checks import is_finite_and_negative_or_nan
from utilities.numpy.checks import is_finite_and_non_negative
from utilities.numpy.checks import is_finite_and_non_negative_or_nan
from utilities.numpy.checks import is_finite_and_non_positive
from utilities.numpy.checks import is_finite_and_non_positive_or_nan
from utilities.numpy.checks import is_finite_and_non_zero
from utilities.numpy.checks import is_finite_and_non_zero_or_nan
from utilities.numpy.checks import is_finite_and_positive
from utilities.numpy.checks import is_finite_and_positive_or_nan
from utilities.numpy.checks import is_finite_or_nan
from utilities.numpy.checks import is_integral
from utilities.numpy.checks import is_integral_or_nan
from utilities.numpy.checks import is_negative
from utilities.numpy.checks import is_negative_or_nan
from utilities.numpy.checks import is_non_negative
from utilities.numpy.checks import is_non_negative_or_nan
from utilities.numpy.checks import is_non_positive
from utilities.numpy.checks import is_non_positive_or_nan
from utilities.numpy.checks import is_non_zero
from utilities.numpy.checks import is_non_zero_or_nan
from utilities.numpy.checks import is_positive
from utilities.numpy.checks import is_positive_or_nan
from utilities.numpy.checks import is_zero
from utilities.numpy.checks import is_zero_or_finite_and_non_micro
from utilities.numpy.checks import is_zero_or_finite_and_non_micro_or_nan
from utilities.numpy.checks import is_zero_or_nan
from utilities.numpy.checks import is_zero_or_non_micro
from utilities.numpy.checks import is_zero_or_non_micro_or_nan
from utilities.numpy.dtypes import datetime64as
from utilities.numpy.dtypes import datetime64D
from utilities.numpy.dtypes import datetime64fs
from utilities.numpy.dtypes import datetime64h
from utilities.numpy.dtypes import datetime64M
from utilities.numpy.dtypes import datetime64m
from utilities.numpy.dtypes import datetime64ms
from utilities.numpy.dtypes import datetime64ns
from utilities.numpy.dtypes import datetime64ps
from utilities.numpy.dtypes import datetime64s
from utilities.numpy.dtypes import datetime64us
from utilities.numpy.dtypes import datetime64W
from utilities.numpy.dtypes import datetime64Y

# dtype checkers
DTypeB = IsAttr["dtype", IsEqual[bool]]
DTypeDY = IsAttr["dtype", IsEqual[datetime64Y]]
DTypeDM = IsAttr["dtype", IsEqual[datetime64M]]
DTypeDW = IsAttr["dtype", IsEqual[datetime64W]]
DTypeDD = IsAttr["dtype", IsEqual[datetime64D]]
DTypeDh = IsAttr["dtype", IsEqual[datetime64h]]
DTypeDm = IsAttr["dtype", IsEqual[datetime64m]]
DTypeDs = IsAttr["dtype", IsEqual[datetime64s]]
DTypeDms = IsAttr["dtype", IsEqual[datetime64ms]]
DTypeDus = IsAttr["dtype", IsEqual[datetime64us]]
DTypeDns = IsAttr["dtype", IsEqual[datetime64ns]]
DTypeDps = IsAttr["dtype", IsEqual[datetime64ps]]
DTypeDfs = IsAttr["dtype", IsEqual[datetime64fs]]
DTypeDas = IsAttr["dtype", IsEqual[datetime64as]]
DTypeD = (
    DTypeDY
    | DTypeDM
    | DTypeDW
    | DTypeDD
    | DTypeDh
    | DTypeDm
    | DTypeDs
    | DTypeDms
    | DTypeDus
    | DTypeDns
    | DTypeDps
    | DTypeDfs
    | DTypeDas
)
DTypeF = IsAttr["dtype", IsEqual[float]]
DTypeI = IsAttr["dtype", IsEqual[int]]
DTypeO = IsAttr["dtype", IsEqual[object]]

# annotated; dtype
NDArrayA = NDArray[Any]
NDArrayB = NDArray[bool_]
NDArrayDY = NDArray[cast(Any, datetime64Y)]
NDArrayDM = NDArray[cast(Any, datetime64M)]
NDArrayDW = NDArray[cast(Any, datetime64W)]
NDArrayDD = NDArray[cast(Any, datetime64D)]
NDArrayDh = NDArray[cast(Any, datetime64h)]
NDArrayDm = NDArray[cast(Any, datetime64m)]
NDArrayDs = NDArray[cast(Any, datetime64s)]
NDArrayDms = NDArray[cast(Any, datetime64ms)]
NDArrayDus = NDArray[cast(Any, datetime64us)]
NDArrayDns = NDArray[cast(Any, datetime64ns)]
NDArrayDps = NDArray[cast(Any, datetime64ps)]
NDArrayDfs = NDArray[cast(Any, datetime64fs)]
NDArrayDas = NDArray[cast(Any, datetime64as)]
NDArrayD = (
    NDArrayDY
    | NDArrayDM
    | NDArrayDW
    | NDArrayDD
    | NDArrayDh
    | NDArrayDm
    | NDArrayDs
    | NDArrayDms
    | NDArrayDus
    | NDArrayDns
    | NDArrayDps
    | NDArrayDfs
    | NDArrayDas
)
NDArrayF = NDArray[float64]
NDArrayI = NDArray[int64]
NDArrayO = NDArray[object_]

# annotated; ndim
NDArray0 = Annotated[NDArrayA, NDim0]
NDArray1 = Annotated[NDArrayA, NDim1]
NDArray2 = Annotated[NDArrayA, NDim2]
NDArray3 = Annotated[NDArrayA, NDim3]

# annotated; dtype & ndim
NDArrayB0 = Annotated[NDArrayB, NDim0]
NDArrayD0 = Annotated[NDArrayD, NDim0]
NDArrayDY0 = Annotated[NDArrayDY, NDim0]
NDArrayDM0 = Annotated[NDArrayDM, NDim0]
NDArrayDW0 = Annotated[NDArrayDW, NDim0]
NDArrayDD0 = Annotated[NDArrayDD, NDim0]
NDArrayDh0 = Annotated[NDArrayDh, NDim0]
NDArrayDm0 = Annotated[NDArrayDm, NDim0]
NDArrayDs0 = Annotated[NDArrayDs, NDim0]
NDArrayDms0 = Annotated[NDArrayDms, NDim0]
NDArrayDus0 = Annotated[NDArrayDus, NDim0]
NDArrayDns0 = Annotated[NDArrayDns, NDim0]
NDArrayDps0 = Annotated[NDArrayDps, NDim0]
NDArrayDfs0 = Annotated[NDArrayDfs, NDim0]
NDArrayDas0 = Annotated[NDArrayDas, NDim0]
NDArrayF0 = Annotated[NDArrayF, NDim0]
NDArrayI0 = Annotated[NDArrayI, NDim0]
NDArrayO0 = Annotated[NDArrayO, NDim0]

NDArrayB1 = Annotated[NDArrayB, NDim1]
NDArrayD1 = Annotated[NDArrayD, NDim1]
NDArrayDY1 = Annotated[NDArrayDY, NDim1]
NDArrayDM1 = Annotated[NDArrayDM, NDim1]
NDArrayDW1 = Annotated[NDArrayDW, NDim1]
NDArrayDD1 = Annotated[NDArrayDD, NDim1]
NDArrayDh1 = Annotated[NDArrayDh, NDim1]
NDArrayDm1 = Annotated[NDArrayDm, NDim1]
NDArrayDs1 = Annotated[NDArrayDs, NDim1]
NDArrayDms1 = Annotated[NDArrayDms, NDim1]
NDArrayDus1 = Annotated[NDArrayDus, NDim1]
NDArrayDns1 = Annotated[NDArrayDns, NDim1]
NDArrayDps1 = Annotated[NDArrayDps, NDim1]
NDArrayDfs1 = Annotated[NDArrayDfs, NDim1]
NDArrayDas1 = Annotated[NDArrayDas, NDim1]
NDArrayF1 = Annotated[NDArrayF, NDim1]
NDArrayI1 = Annotated[NDArrayI, NDim1]
NDArrayO1 = Annotated[NDArrayO, NDim1]

NDArrayB2 = Annotated[NDArrayB, NDim2]
NDArrayD2 = Annotated[NDArrayD, NDim2]
NDArrayDY2 = Annotated[NDArrayDY, NDim2]
NDArrayDM2 = Annotated[NDArrayDM, NDim2]
NDArrayDW2 = Annotated[NDArrayDW, NDim2]
NDArrayDD2 = Annotated[NDArrayDD, NDim2]
NDArrayDh2 = Annotated[NDArrayDh, NDim2]
NDArrayDm2 = Annotated[NDArrayDm, NDim2]
NDArrayDs2 = Annotated[NDArrayDs, NDim2]
NDArrayDms2 = Annotated[NDArrayDms, NDim2]
NDArrayDus2 = Annotated[NDArrayDus, NDim2]
NDArrayDns2 = Annotated[NDArrayDns, NDim2]
NDArrayDps2 = Annotated[NDArrayDps, NDim2]
NDArrayDfs2 = Annotated[NDArrayDfs, NDim2]
NDArrayDas2 = Annotated[NDArrayDas, NDim2]
NDArrayF2 = Annotated[NDArrayF, NDim2]
NDArrayI2 = Annotated[NDArrayI, NDim2]
NDArrayO2 = Annotated[NDArrayO, NDim2]

NDArrayB3 = Annotated[NDArrayB, NDim3]
NDArrayD3 = Annotated[NDArrayD, NDim3]
NDArrayDY3 = Annotated[NDArrayDY, NDim3]
NDArrayDM3 = Annotated[NDArrayDM, NDim3]
NDArrayDW3 = Annotated[NDArrayDW, NDim3]
NDArrayDD3 = Annotated[NDArrayDD, NDim3]
NDArrayDh3 = Annotated[NDArrayDh, NDim3]
NDArrayDm3 = Annotated[NDArrayDm, NDim3]
NDArrayDs3 = Annotated[NDArrayDs, NDim3]
NDArrayDms3 = Annotated[NDArrayDms, NDim3]
NDArrayDus3 = Annotated[NDArrayDus, NDim3]
NDArrayDns3 = Annotated[NDArrayDns, NDim3]
NDArrayDps3 = Annotated[NDArrayDps, NDim3]
NDArrayDfs3 = Annotated[NDArrayDfs, NDim3]
NDArrayDas3 = Annotated[NDArrayDas, NDim3]
NDArrayF3 = Annotated[NDArrayF, NDim3]
NDArrayI3 = Annotated[NDArrayI, NDim3]
NDArrayO3 = Annotated[NDArrayO, NDim3]


def _lift(check: Callable[..., Any], /) -> Any:
    """Lift a check to work on a subset of a float array."""
    rng = default_rng()

    def predicate(array: NDArrayI | NDArrayF, /) -> bool:
        if (size := array.size) == 0:
            return True
        if size == 1:
            return check(array).item()
        num_samples = round(log(size))
        indices = rng.integers(0, size, size=num_samples)
        sample = array[unravel_index(indices, array.shape)]
        return check(sample).all().item()

    return Is[cast(Any, predicate)]


_is_finite = _lift(isfinite)
_is_finite_and_integral = _lift(is_finite_and_integral)
_is_finite_and_integral_or_nan = _lift(is_finite_and_integral_or_nan)
_is_finite_and_negative = _lift(is_finite_and_negative)
_is_finite_and_negative_or_nan = _lift(is_finite_and_negative_or_nan)
_is_finite_and_non_negative = _lift(is_finite_and_non_negative)
_is_finite_and_non_negative_or_nan = _lift(is_finite_and_non_negative_or_nan)
_is_finite_and_non_positive = _lift(is_finite_and_non_positive)
_is_finite_and_non_positive_or_nan = _lift(is_finite_and_non_positive_or_nan)
_is_finite_and_non_zero = _lift(is_finite_and_non_zero)
_is_finite_and_non_zero_or_nan = _lift(is_finite_and_non_zero_or_nan)
_is_finite_and_positive = _lift(is_finite_and_positive)
_is_finite_and_positive_or_nan = _lift(is_finite_and_positive_or_nan)
_is_finite_or_nan = _lift(is_finite_or_nan)
_is_integral = _lift(is_integral)
_is_integral_or_nan = _lift(is_integral_or_nan)
_is_negative = _lift(is_negative)
_is_negative_or_nan = _lift(is_negative_or_nan)
_is_non_negative = _lift(is_non_negative)
_is_non_negative_or_nan = _lift(is_non_negative_or_nan)
_is_non_positive = _lift(is_non_positive)
_is_non_positive_or_nan = _lift(is_non_positive_or_nan)
_is_non_zero = _lift(is_non_zero)
_is_non_zero_or_nan = _lift(is_non_zero_or_nan)
_is_positive = _lift(is_positive)
_is_positive_or_nan = _lift(is_positive_or_nan)
_is_zero = _lift(is_zero)
_is_zero_or_finite_and_non_micro = _lift(is_zero_or_finite_and_non_micro)
_is_zero_or_finite_and_non_micro_or_nan = _lift(
    is_zero_or_finite_and_non_micro_or_nan
)
_is_zero_or_nan = _lift(is_zero_or_nan)
_is_zero_or_non_micro = _lift(is_zero_or_non_micro)
_is_zero_or_non_micro_or_nan = _lift(is_zero_or_non_micro_or_nan)


# annotated; int & checks
NDArrayINeg = Annotated[NDArrayI, _is_negative]
NDArrayINonNeg = Annotated[NDArrayI, _is_non_negative]
NDArrayINonPos = Annotated[NDArrayI, _is_non_positive]
NDArrayINonZr = Annotated[NDArrayI, _is_non_zero]
NDArrayIPos = Annotated[NDArrayI, _is_positive]
NDArrayIZr = Annotated[NDArrayI, _is_zero]


# annotated; float & checks
NDArrayFFin = Annotated[NDArrayF, _is_finite]
NDArrayFFinInt = Annotated[NDArrayF, _is_finite_and_integral]
NDArrayFFinIntNan = Annotated[NDArrayF, _is_finite_and_integral_or_nan]
NDArrayFFinNeg = Annotated[NDArrayF, _is_finite_and_negative]
NDArrayFFinNegNan = Annotated[NDArrayF, _is_finite_and_negative_or_nan]
NDArrayFFinNonNeg = Annotated[NDArrayF, _is_finite_and_non_negative]
NDArrayFFinNonNegNan = Annotated[NDArrayF, _is_finite_and_non_negative_or_nan]
NDArrayFFinNonPos = Annotated[NDArrayF, _is_finite_and_non_positive]
NDArrayFFinNonPosNan = Annotated[NDArrayF, _is_finite_and_non_positive_or_nan]
NDArrayFFinNonZr = Annotated[NDArrayF, _is_finite_and_non_zero]
NDArrayFFinNonZrNan = Annotated[NDArrayF, _is_finite_and_non_zero_or_nan]
NDArrayFFinPos = Annotated[NDArrayF, _is_finite_and_positive]
NDArrayFFinPosNan = Annotated[NDArrayF, _is_finite_and_positive_or_nan]
NDArrayFFinNan = Annotated[NDArrayF, _is_finite_or_nan]
NDArrayFInt = Annotated[NDArrayF, _is_integral]
NDArrayFIntNan = Annotated[NDArrayF, _is_integral_or_nan]
NDArrayFNeg = Annotated[NDArrayF, _is_negative]
NDArrayFNegNan = Annotated[NDArrayF, _is_negative_or_nan]
NDArrayFNonNeg = Annotated[NDArrayF, _is_non_negative]
NDArrayFNonNegNan = Annotated[NDArrayF, _is_non_negative_or_nan]
NDArrayFNonPos = Annotated[NDArrayF, _is_non_positive]
NDArrayFNonPosNan = Annotated[NDArrayF, _is_non_positive_or_nan]
NDArrayFNonZr = Annotated[NDArrayF, _is_non_zero]
NDArrayFNonZrNan = Annotated[NDArrayF, _is_non_zero_or_nan]
NDArrayFPos = Annotated[NDArrayF, _is_positive]
NDArrayFPosNan = Annotated[NDArrayF, _is_positive_or_nan]
NDArrayFZr = Annotated[NDArrayF, _is_zero]
NDArrayFZrFinNonMic = Annotated[NDArrayF, _is_zero_or_finite_and_non_micro]
NDArrayFZrFinNonMicNan = Annotated[
    NDArrayF, _is_zero_or_finite_and_non_micro_or_nan
]
NDArrayFZrNan = Annotated[NDArrayF, _is_zero_or_nan]
NDArrayFZrNonMic = Annotated[NDArrayF, _is_zero_or_non_micro]
NDArrayFZrNonMicNan = Annotated[NDArrayF, _is_zero_or_non_micro_or_nan]

# annotated; int, ndim & checks
NDArrayI0Neg = Annotated[NDArrayI, NDim0 & _is_negative]
NDArrayI0NonNeg = Annotated[NDArrayI, NDim0 & _is_non_negative]
NDArrayI0NonPos = Annotated[NDArrayI, NDim0 & _is_non_positive]
NDArrayI0NonZr = Annotated[NDArrayI, NDim0 & _is_non_zero]
NDArrayI0Pos = Annotated[NDArrayI, NDim0 & _is_positive]
NDArrayI0Zr = Annotated[NDArrayI, NDim0 & _is_zero]

NDArrayI1Neg = Annotated[NDArrayI, NDim1 & _is_negative]
NDArrayI1NonNeg = Annotated[NDArrayI, NDim1 & _is_non_negative]
NDArrayI1NonPos = Annotated[NDArrayI, NDim1 & _is_non_positive]
NDArrayI1NonZr = Annotated[NDArrayI, NDim1 & _is_non_zero]
NDArrayI1Pos = Annotated[NDArrayI, NDim1 & _is_positive]
NDArrayI1Zr = Annotated[NDArrayI, NDim1 & _is_zero]

NDArrayI2Neg = Annotated[NDArrayI, NDim2 & _is_negative]
NDArrayI2NonNeg = Annotated[NDArrayI, NDim2 & _is_non_negative]
NDArrayI2NonPos = Annotated[NDArrayI, NDim2 & _is_non_positive]
NDArrayI2NonZr = Annotated[NDArrayI, NDim2 & _is_non_zero]
NDArrayI2Pos = Annotated[NDArrayI, NDim2 & _is_positive]
NDArrayI2Zr = Annotated[NDArrayI, NDim2 & _is_zero]

NDArrayI3Neg = Annotated[NDArrayI, NDim1 & _is_negative]
NDArrayI3NonNeg = Annotated[NDArrayI, NDim3 & _is_non_negative]
NDArrayI3NonPos = Annotated[NDArrayI, NDim3 & _is_non_positive]
NDArrayI3NonZr = Annotated[NDArrayI, NDim3 & _is_non_zero]
NDArrayI3Pos = Annotated[NDArrayI, NDim3 & _is_positive]
NDArrayI3Zr = Annotated[NDArrayI, NDim3 & _is_zero]

# annotated; float, ndim & checks
NDArrayF0Fin = Annotated[NDArrayF, NDim0 & _is_finite]
NDArrayF0FinInt = Annotated[NDArrayF, NDim0 & _is_finite_and_integral]
NDArrayF0FinIntNan = Annotated[NDArrayF, NDim0 & _is_finite_and_integral_or_nan]
NDArrayF0FinNeg = Annotated[NDArrayF, NDim0 & _is_finite_and_negative]
NDArrayF0FinNegNan = Annotated[NDArrayF, NDim0 & _is_finite_and_negative_or_nan]
NDArrayF0FinNonNeg = Annotated[NDArrayF, NDim0 & _is_finite_and_non_negative]
NDArrayF0FinNonNegNan = Annotated[
    NDArrayF, NDim0 & _is_finite_and_non_negative_or_nan
]
NDArrayF0FinNonPos = Annotated[NDArrayF, NDim0 & _is_finite_and_non_positive]
NDArrayF0FinNonPosNan = Annotated[
    NDArrayF, NDim0 & _is_finite_and_non_positive_or_nan
]
NDArrayF0FinNonZr = Annotated[NDArrayF, NDim0 & _is_finite_and_non_zero]
NDArrayF0FinNonZrNan = Annotated[
    NDArrayF, NDim0 & _is_finite_and_non_zero_or_nan
]
NDArrayF0FinPos = Annotated[NDArrayF, NDim0 & _is_finite_and_positive]
NDArrayF0FinPosNan = Annotated[NDArrayF, NDim0 & _is_finite_and_positive_or_nan]
NDArrayF0FinNan = Annotated[NDArrayF, NDim0 & _is_finite_or_nan]
NDArrayF0Int = Annotated[NDArrayF, NDim0 & _is_integral]
NDArrayF0IntNan = Annotated[NDArrayF, NDim0 & _is_integral_or_nan]
NDArrayF0Neg = Annotated[NDArrayF, NDim0 & _is_negative]
NDArrayF0NegNan = Annotated[NDArrayF, NDim0 & _is_negative_or_nan]
NDArrayF0NonNeg = Annotated[NDArrayF, NDim0 & _is_non_negative]
NDArrayF0NonNegNan = Annotated[NDArrayF, NDim0 & _is_non_negative_or_nan]
NDArrayF0NonPos = Annotated[NDArrayF, NDim0 & _is_non_positive]
NDArrayF0NonPosNan = Annotated[NDArrayF, NDim0 & _is_non_positive_or_nan]
NDArrayF0NonZr = Annotated[NDArrayF, NDim0 & _is_non_zero]
NDArrayF0NonZrNan = Annotated[NDArrayF, NDim0 & _is_non_zero_or_nan]
NDArrayF0Pos = Annotated[NDArrayF, NDim0 & _is_positive]
NDArrayF0PosNan = Annotated[NDArrayF, NDim0 & _is_positive_or_nan]
NDArrayF0Zr = Annotated[NDArrayF, NDim0 & _is_zero]
NDArrayF0ZrFinNonMic = Annotated[
    NDArrayF, NDim0 & _is_zero_or_finite_and_non_micro
]
NDArrayF0ZrFinNonMicNan = Annotated[
    NDArrayF, NDim0 & _is_zero_or_finite_and_non_micro_or_nan
]
NDArrayF0ZrNan = Annotated[NDArrayF, NDim0 & _is_zero_or_nan]
NDArrayF0ZrNonMic = Annotated[NDArrayF, NDim0 & _is_zero_or_non_micro]
NDArrayF0ZrNonMicNan = Annotated[NDArrayF, NDim0 & _is_zero_or_non_micro_or_nan]

NDArrayF1Fin = Annotated[NDArrayF, NDim1 & _is_finite]
NDArrayF1FinInt = Annotated[NDArrayF, NDim1 & _is_finite_and_integral]
NDArrayF1FinIntNan = Annotated[NDArrayF, NDim1 & _is_finite_and_integral_or_nan]
NDArrayF1FinNeg = Annotated[NDArrayF, NDim1 & _is_finite_and_negative]
NDArrayF1FinNegNan = Annotated[NDArrayF, NDim1 & _is_finite_and_negative_or_nan]
NDArrayF1FinNonNeg = Annotated[NDArrayF, NDim1 & _is_finite_and_non_negative]
NDArrayF1FinNonNegNan = Annotated[
    NDArrayF, NDim1 & _is_finite_and_non_negative_or_nan
]
NDArrayF1FinNonPos = Annotated[NDArrayF, NDim1 & _is_finite_and_non_positive]
NDArrayF1FinNonPosNan = Annotated[
    NDArrayF, NDim1 & _is_finite_and_non_positive_or_nan
]
NDArrayF1FinNonZr = Annotated[NDArrayF, NDim1 & _is_finite_and_non_zero]
NDArrayF1FinNonZrNan = Annotated[
    NDArrayF, NDim1 & _is_finite_and_non_zero_or_nan
]
NDArrayF1FinPos = Annotated[NDArrayF, NDim1 & _is_finite_and_positive]
NDArrayF1FinPosNan = Annotated[NDArrayF, NDim1 & _is_finite_and_positive_or_nan]
NDArrayF1FinNan = Annotated[NDArrayF, NDim1 & _is_finite_or_nan]
NDArrayF1Int = Annotated[NDArrayF, NDim1 & _is_integral]
NDArrayF1IntNan = Annotated[NDArrayF, NDim1 & _is_integral_or_nan]
NDArrayF1Neg = Annotated[NDArrayF, NDim1 & _is_negative]
NDArrayF1NegNan = Annotated[NDArrayF, NDim1 & _is_negative_or_nan]
NDArrayF1NonNeg = Annotated[NDArrayF, NDim1 & _is_non_negative]
NDArrayF1NonNegNan = Annotated[NDArrayF, NDim1 & _is_non_negative_or_nan]
NDArrayF1NonPos = Annotated[NDArrayF, NDim1 & _is_non_positive]
NDArrayF1NonPosNan = Annotated[NDArrayF, NDim1 & _is_non_positive_or_nan]
NDArrayF1NonZr = Annotated[NDArrayF, NDim1 & _is_non_zero]
NDArrayF1NonZrNan = Annotated[NDArrayF, NDim1 & _is_non_zero_or_nan]
NDArrayF1Pos = Annotated[NDArrayF, NDim1 & _is_positive]
NDArrayF1PosNan = Annotated[NDArrayF, NDim1 & _is_positive_or_nan]
NDArrayF1Zr = Annotated[NDArrayF, NDim1 & _is_zero]
NDArrayF1ZrFinNonMic = Annotated[
    NDArrayF, NDim1 & _is_zero_or_finite_and_non_micro
]
NDArrayF1ZrFinNonMicNan = Annotated[
    NDArrayF, NDim1 & _is_zero_or_finite_and_non_micro_or_nan
]
NDArrayF1ZrNan = Annotated[NDArrayF, NDim1 & _is_zero_or_nan]
NDArrayF1ZrNonMic = Annotated[NDArrayF, NDim1 & _is_zero_or_non_micro]
NDArrayF1ZrNonMicNan = Annotated[NDArrayF, NDim1 & _is_zero_or_non_micro_or_nan]

NDArrayF2Fin = Annotated[NDArrayF, NDim2 & _is_finite]
NDArrayF2FinInt = Annotated[NDArrayF, NDim2 & _is_finite_and_integral]
NDArrayF2FinIntNan = Annotated[NDArrayF, NDim2 & _is_finite_and_integral_or_nan]
NDArrayF2FinNeg = Annotated[NDArrayF, NDim2 & _is_finite_and_negative]
NDArrayF2FinNegNan = Annotated[NDArrayF, NDim2 & _is_finite_and_negative_or_nan]
NDArrayF2FinNonNeg = Annotated[NDArrayF, NDim2 & _is_finite_and_non_negative]
NDArrayF2FinNonNegNan = Annotated[
    NDArrayF, NDim2 & _is_finite_and_non_negative_or_nan
]
NDArrayF2FinNonPos = Annotated[NDArrayF, NDim2 & _is_finite_and_non_positive]
NDArrayF2FinNonPosNan = Annotated[
    NDArrayF, NDim2 & _is_finite_and_non_positive_or_nan
]
NDArrayF2FinNonZr = Annotated[NDArrayF, NDim2 & _is_finite_and_non_zero]
NDArrayF2FinNonZrNan = Annotated[
    NDArrayF, NDim2 & _is_finite_and_non_zero_or_nan
]
NDArrayF2FinPos = Annotated[NDArrayF, NDim2 & _is_finite_and_positive]
NDArrayF2FinPosNan = Annotated[NDArrayF, NDim2 & _is_finite_and_positive_or_nan]
NDArrayF2FinNan = Annotated[NDArrayF, NDim2 & _is_finite_or_nan]
NDArrayF2Int = Annotated[NDArrayF, NDim2 & _is_integral]
NDArrayF2IntNan = Annotated[NDArrayF, NDim2 & _is_integral_or_nan]
NDArrayF2Neg = Annotated[NDArrayF, NDim2 & _is_negative]
NDArrayF2NegNan = Annotated[NDArrayF, NDim2 & _is_negative_or_nan]
NDArrayF2NonNeg = Annotated[NDArrayF, NDim2 & _is_non_negative]
NDArrayF2NonNegNan = Annotated[NDArrayF, NDim2 & _is_non_negative_or_nan]
NDArrayF2NonPos = Annotated[NDArrayF, NDim2 & _is_non_positive]
NDArrayF2NonPosNan = Annotated[NDArrayF, NDim2 & _is_non_positive_or_nan]
NDArrayF2NonZr = Annotated[NDArrayF, NDim2 & _is_non_zero]
NDArrayF2NonZrNan = Annotated[NDArrayF, NDim2 & _is_non_zero_or_nan]
NDArrayF2Pos = Annotated[NDArrayF, NDim2 & _is_positive]
NDArrayF2PosNan = Annotated[NDArrayF, NDim2 & _is_positive_or_nan]
NDArrayF2Zr = Annotated[NDArrayF, NDim2 & _is_zero]
NDArrayF2ZrFinNonMic = Annotated[
    NDArrayF, NDim2 & _is_zero_or_finite_and_non_micro
]
NDArrayF2ZrFinNonMicNan = Annotated[
    NDArrayF, NDim2 & _is_zero_or_finite_and_non_micro_or_nan
]
NDArrayF2ZrNan = Annotated[NDArrayF, NDim2 & _is_zero_or_nan]
NDArrayF2ZrNonMic = Annotated[NDArrayF, NDim2 & _is_zero_or_non_micro]
NDArrayF2ZrNonMicNan = Annotated[NDArrayF, NDim2 & _is_zero_or_non_micro_or_nan]

NDArrayF3Fin = Annotated[NDArrayF, NDim3 & _is_finite]
NDArrayF3FinInt = Annotated[NDArrayF, NDim3 & _is_finite_and_integral]
NDArrayF3FinIntNan = Annotated[NDArrayF, NDim3 & _is_finite_and_integral_or_nan]
NDArrayF3FinNeg = Annotated[NDArrayF, NDim3 & _is_finite_and_negative]
NDArrayF3FinNegNan = Annotated[NDArrayF, NDim3 & _is_finite_and_negative_or_nan]
NDArrayF3FinNonNeg = Annotated[NDArrayF, NDim3 & _is_finite_and_non_negative]
NDArrayF3FinNonNegNan = Annotated[
    NDArrayF, NDim3 & _is_finite_and_non_negative_or_nan
]
NDArrayF3FinNonPos = Annotated[NDArrayF, NDim3 & _is_finite_and_non_positive]
NDArrayF3FinNonPosNan = Annotated[
    NDArrayF, NDim3 & _is_finite_and_non_positive_or_nan
]
NDArrayF3FinNonZr = Annotated[NDArrayF, NDim3 & _is_finite_and_non_zero]
NDArrayF3FinNonZrNan = Annotated[
    NDArrayF, NDim3 & _is_finite_and_non_zero_or_nan
]
NDArrayF3FinPos = Annotated[NDArrayF, NDim3 & _is_finite_and_positive]
NDArrayF3FinPosNan = Annotated[NDArrayF, NDim3 & _is_finite_and_positive_or_nan]
NDArrayF3FinNan = Annotated[NDArrayF, NDim3 & _is_finite_or_nan]
NDArrayF3Int = Annotated[NDArrayF, NDim3 & _is_integral]
NDArrayF3IntNan = Annotated[NDArrayF, NDim3 & _is_integral_or_nan]
NDArrayF3Neg = Annotated[NDArrayF, NDim3 & _is_negative]
NDArrayF3NegNan = Annotated[NDArrayF, NDim3 & _is_negative_or_nan]
NDArrayF3NonNeg = Annotated[NDArrayF, NDim3 & _is_non_negative]
NDArrayF3NonNegNan = Annotated[NDArrayF, NDim3 & _is_non_negative_or_nan]
NDArrayF3NonPos = Annotated[NDArrayF, NDim3 & _is_non_positive]
NDArrayF3NonPosNan = Annotated[NDArrayF, NDim3 & _is_non_positive_or_nan]
NDArrayF3NonZr = Annotated[NDArrayF, NDim3 & _is_non_zero]
NDArrayF3NonZrNan = Annotated[NDArrayF, NDim3 & _is_non_zero_or_nan]
NDArrayF3Pos = Annotated[NDArrayF, NDim3 & _is_positive]
NDArrayF3PosNan = Annotated[NDArrayF, NDim3 & _is_positive_or_nan]
NDArrayF3Zr = Annotated[NDArrayF, NDim3 & _is_zero]
NDArrayF3ZrFinNonMic = Annotated[
    NDArrayF, NDim3 & _is_zero_or_finite_and_non_micro
]
NDArrayF3ZrFinNonMicNan = Annotated[
    NDArrayF, NDim3 & _is_zero_or_finite_and_non_micro_or_nan
]
NDArrayF3ZrNan = Annotated[NDArrayF, NDim3 & _is_zero_or_nan]
NDArrayF3ZrNonMic = Annotated[NDArrayF, NDim3 & _is_zero_or_non_micro]
NDArrayF3ZrNonMicNan = Annotated[NDArrayF, NDim3 & _is_zero_or_non_micro_or_nan]



from __future__ import annotations

import datetime as dt
from collections.abc import Set as AbstractSet
from math import inf
from math import isfinite
from math import isinf
from math import isnan
from pathlib import Path
from re import search
from subprocess import PIPE
from subprocess import check_output

from hypothesis import Phase
from hypothesis import assume
from hypothesis import given
from hypothesis import settings
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import DataObject
from hypothesis.strategies import DrawFn
from hypothesis.strategies import booleans
from hypothesis.strategies import composite
from hypothesis.strategies import data
from hypothesis.strategies import datetimes
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import none
from hypothesis.strategies import sets
from more_itertools import pairwise
from pytest import mark
from pytest import param
from pytest import raises

from utilities.datetime import UTC
from utilities.git import _GET_BRANCH_NAME
from utilities.hypothesis import _MAX_EXAMPLES
from utilities.hypothesis import _NO_SHRINK
from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import datetimes_utc
from utilities.hypothesis import floats_extra
from utilities.hypothesis import git_repos
from utilities.hypothesis import hashables
from utilities.hypothesis import lists_fixed_length
from utilities.hypothesis import setup_hypothesis_profiles
from utilities.hypothesis import slices
from utilities.hypothesis import temp_dirs
from utilities.hypothesis import temp_paths
from utilities.hypothesis import text_ascii
from utilities.hypothesis import text_clean
from utilities.hypothesis import text_printable
from utilities.os import temp_environ
from utilities.platform import maybe_yield_lower_case
from utilities.tempfile import TemporaryDirectory
from utilities.typing import IterableStrs


class TestAssumeDoesNotRaise:
    @given(x=booleans())
    def test_no_match_and_suppressed(self, x: bool) -> None:
        with assume_does_not_raise(ValueError):
            if x is True:
                msg = "x is True"
                raise ValueError(msg)
        assert x is False

    @given(x=booleans())
    def test_no_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                RuntimeError
            ):
                raise ValueError(msg)

    @given(x=booleans())
    def test_with_match_and_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with assume_does_not_raise(ValueError, match=msg):
                raise ValueError(msg)
        assert x is False

    @given(x=just(True))
    def test_with_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                ValueError, match="wrong"
            ):
                raise ValueError(msg)


class TestDatetimesUTC:
    @given(data=data(), min_value=datetimes(), max_value=datetimes())
    def test_main(
        self, data: DataObject, min_value: dt.datetime, max_value: dt.datetime
    ) -> None:
        min_value, max_value = (
            v.replace(tzinfo=UTC) for v in [min_value, max_value]
        )
        _ = assume(min_value <= max_value)
        datetime = data.draw(
            datetimes_utc(min_value=min_value, max_value=max_value)
        )
        assert min_value <= datetime <= max_value


class TestFloatsExtra:
    @given(
        data=data(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans(),
        allow_inf=booleans(),
        allow_pos_inf=booleans(),
        allow_neg_inf=booleans(),
        integral=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_value: float | None,
        max_value: float | None,
        allow_nan: bool,
        allow_inf: bool,
        allow_pos_inf: bool,
        allow_neg_inf: bool,
        integral: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(
                floats_extra(
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_inf=allow_inf,
                    allow_pos_inf=allow_pos_inf,
                    allow_neg_inf=allow_neg_inf,
                    integral=integral,
                )
            )
        if min_value is not None:
            assert (isfinite(x) and x >= min_value) or not isfinite(x)
        if max_value is not None:
            assert (isfinite(x) and x <= max_value) or not isfinite(x)
        if not allow_nan:
            assert not isnan(x)
        if not allow_inf:
            if not (allow_pos_inf or allow_neg_inf):
                assert not isinf(x)
            if not allow_pos_inf:
                assert x != inf
            if not allow_neg_inf:
                assert x != -inf
        if integral:
            assert (isfinite(x) and x == round(x)) or not isfinite(x)


class TestGitRepos:
    @given(data=data())
    def test_fixed(self, *, data: DataObject) -> None:
        branch = data.draw(text_ascii(min_size=1) | none())
        repo = data.draw(git_repos(branch=branch))
        assert isinstance(repo, TemporaryDirectory)
        path = repo.path
        assert set(path.iterdir()) == {path.joinpath(".git")}
        if branch is not None:
            output = check_output(
                _GET_BRANCH_NAME, stderr=PIPE, cwd=path, text=True  # noqa: S603
            )
            assert output.strip("\n") == branch


class TestHashables:
    @given(data=data())
    def test_fixed(self, data: DataObject) -> None:
        x = data.draw(hashables())
        _ = hash(x)


class TestLiftDraw:
    @given(data=data(), x=booleans())
    def test_fixed(self, data: DataObject, x: bool) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            _ = _draw(booleans())
            return x

        result = data.draw(func())
        assert result is x

    @given(data=data())
    def test_strategy(self, data: DataObject) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            return _draw(booleans())

        result = data.draw(func())
        assert isinstance(result, bool)


class TestListsFixedLength:
    @given(data=data(), size=integers(1, 10))
    @mark.parametrize(
        "unique", [param(True, id="unique"), param(False, id="no unique")]
    )
    @mark.parametrize(
        "sorted_", [param(True, id="sorted"), param(False, id="no sorted")]
    )
    def test_main(
        self, data: DataObject, size: int, unique: bool, sorted_: bool
    ) -> None:
        result = data.draw(
            lists_fixed_length(integers(), size, unique=unique, sorted=sorted_)
        )
        assert isinstance(result, list)
        assert len(result) == size
        if unique:
            assert len(set(result)) == len(result)
        if sorted_:
            assert sorted(result) == result


class TestSlices:
    @given(data=data(), iter_len=integers(0, 10))
    def test_main(self, data: DataObject, iter_len: int) -> None:
        slice_len = data.draw(integers(0, iter_len) | none())
        slice_ = data.draw(slices(iter_len, slice_len=slice_len))
        range_slice = range(iter_len)[slice_]
        assert all(i + 1 == j for i, j in pairwise(range_slice))
        if slice_len is not None:
            assert len(range_slice) == slice_len

    @given(data=data(), iter_len=integers(0, 10))
    def test_error(self, data: DataObject, iter_len: int) -> None:
        with raises(
            InvalidArgument,
            match=r"Slice length \d+ exceeds iterable length \d+",
        ):
            _ = data.draw(slices(iter_len, slice_len=iter_len + 1))


class TestSetupHypothesisProfiles:
    def test_main(self) -> None:
        setup_hypothesis_profiles()
        curr = settings()
        assert Phase.shrink in curr.phases
        assert curr.max_examples in {10, 100, 1000}

    def test_no_shrink(self) -> None:
        with temp_environ({_NO_SHRINK: "1"}):
            setup_hypothesis_profiles()
        assert Phase.shrink not in settings().phases

    @given(max_examples=integers(1, 100))
    def test_max_examples(self, max_examples: int) -> None:
        with temp_environ({_MAX_EXAMPLES: str(max_examples)}):
            setup_hypothesis_profiles()
        assert settings().max_examples == max_examples


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, *, temp_dir: TemporaryDirectory) -> None:
        _test_temp_path(temp_dir.path)

    @given(
        temp_dir=temp_dirs(), contents=sets(text_ascii(min_size=1), max_size=10)
    )
    def test_writing_files(
        self, *, temp_dir: TemporaryDirectory, contents: AbstractSet[str]
    ) -> None:
        _test_writing_to_temp_path(temp_dir.path, contents)


class TestTempPaths:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        _test_temp_path(temp_path)

    @given(
        temp_path=temp_paths(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(
        self, *, temp_path: Path, contents: AbstractSet[str]
    ) -> None:
        _test_writing_to_temp_path(temp_path, contents)


def _test_temp_path(path: Path, /) -> None:
    assert path.is_dir()
    assert len(set(path.iterdir())) == 0


def _test_writing_to_temp_path(path: Path, contents: IterableStrs, /) -> None:
    assert len(set(path.iterdir())) == 0
    as_set = set(maybe_yield_lower_case(contents))
    for content in as_set:
        path.joinpath(content).touch()
    assert len(set(path.iterdir())) == len(as_set)


class TestTextAscii:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_ascii(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                )
            )
        assert search("^[A-Za-z]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestTextClean:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_clean(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                )
            )
        assert search("^\\S[^\\r\\n]*$|^$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestTextPrintable:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_printable(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                )
            )
        assert search(
            r"^[0-9A-Za-z!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~\s]*$", text
        )
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"



from __future__ import annotations

from re import search

from hypothesis import given
from pytest import mark

from utilities.hypothesis import text_clean
from utilities.platform import SYSTEM
from utilities.platform import System
from utilities.platform.datetime import maybe_sub_pct_y


class TestMaybeMaybeSubPctY:
    @given(text=text_clean())
    @mark.skipif(SYSTEM is not System.linux, reason="Linux only")
    def test_main(self, text: str) -> None:
        result = maybe_sub_pct_y(text)
        assert not search("%Y", result)



from __future__ import annotations

from typing import TypeVar

from hypothesis.strategies import SearchStrategy

_T = TypeVar("_T")
MaybeSearchStrategy = _T | SearchStrategy[_T]


Shape = int | tuple[int, ...]



from __future__ import annotations

from typing import Annotated
from typing import TypeAlias

from xarray import DataArray

from utilities.beartype import NDim0
from utilities.beartype import NDim1
from utilities.beartype import NDim2
from utilities.beartype import NDim3
from utilities.numpy.typing import DTypeB
from utilities.numpy.typing import DTypeDns
from utilities.numpy.typing import DTypeF
from utilities.numpy.typing import DTypeI
from utilities.numpy.typing import DTypeO

# dtype annotated;
DataArrayB: TypeAlias = Annotated[DataArray, DTypeB]
DataArrayDns: TypeAlias = Annotated[DataArray, DTypeDns]
DataArrayF: TypeAlias = Annotated[DataArray, DTypeF]
DataArrayI: TypeAlias = Annotated[DataArray, DTypeI]
DataArrayO: TypeAlias = Annotated[DataArray, DTypeO]

# ndim annotated;
DataArray0: TypeAlias = Annotated[DataArray, NDim0]
DataArray1: TypeAlias = Annotated[DataArray, NDim1]
DataArray2: TypeAlias = Annotated[DataArray, NDim2]
DataArray3: TypeAlias = Annotated[DataArray, NDim3]

# annotated; dtype & ndim
DataArrayB0: TypeAlias = Annotated[DataArray, DTypeB & NDim0]
DataArrayDns0: TypeAlias = Annotated[DataArray, DTypeDns & NDim0]
DataArrayF0: TypeAlias = Annotated[DataArray, DTypeF & NDim0]
DataArrayI0: TypeAlias = Annotated[DataArray, DTypeI & NDim0]
DataArrayO0: TypeAlias = Annotated[DataArray, DTypeO & NDim0]

DataArrayB1: TypeAlias = Annotated[DataArray, DTypeB & NDim1]
DataArrayDns1: TypeAlias = Annotated[DataArray, DTypeDns & NDim1]
DataArrayF1: TypeAlias = Annotated[DataArray, DTypeF & NDim1]
DataArrayI1: TypeAlias = Annotated[DataArray, DTypeI & NDim1]
DataArrayO1: TypeAlias = Annotated[DataArray, DTypeO & NDim1]

DataArrayB2: TypeAlias = Annotated[DataArray, DTypeB & NDim2]
DataArrayDns2: TypeAlias = Annotated[DataArray, DTypeDns & NDim2]
DataArrayF2: TypeAlias = Annotated[DataArray, DTypeF & NDim2]
DataArrayI2: TypeAlias = Annotated[DataArray, DTypeI & NDim2]
DataArrayO2: TypeAlias = Annotated[DataArray, DTypeO & NDim2]

DataArrayB3: TypeAlias = Annotated[DataArray, DTypeB & NDim3]
DataArrayDns3: TypeAlias = Annotated[DataArray, DTypeDns & NDim3]
DataArrayF3: TypeAlias = Annotated[DataArray, DTypeF & NDim3]
DataArrayI3: TypeAlias = Annotated[DataArray, DTypeI & NDim3]
DataArrayO3: TypeAlias = Annotated[DataArray, DTypeO & NDim3]



from __future__ import annotations

import numpy as np
from numba import float32, float64

from utilities._numbagg.decorators import ndmovingexp


@ndmovingexp(
    [(float32[:], float32, float32[:]), (float64[:], float64, float64[:])]
)
def move_exp_nanmean(a, alpha, out):
    # Inspired by pandas:
    # https://github.com/pandas-dev/pandas/blob/1.2.x/pandas/_libs/window/aggregations.pyx#L1559.

    N = len(a)
    if N == 0:
        return

    old_wt_factor = 1.0 - alpha
    new_wt = 1.0
    ignore_na = False  # could add as option in the future

    weighted_avg = np.nan
    n_obs = 0
    old_wt = 1.0

    for i in range(N):
        cur = a[i]
        is_observation = not np.isnan(cur)
        n_obs += int(is_observation)
        if not np.isnan(weighted_avg):
            if is_observation or (not ignore_na):
                old_wt *= old_wt_factor

                if is_observation:
                    # avoid numerical errors on constant series
                    if weighted_avg != cur:
                        weighted_avg = (
                            (old_wt * weighted_avg) + (new_wt * cur)
                        ) / (old_wt + new_wt)
                    old_wt += new_wt
        elif is_observation:
            # The first non-nan value.
            weighted_avg = cur

        out[i] = weighted_avg


@ndmovingexp(
    [(float32[:], float32, float32[:]), (float64[:], float64, float64[:])]
)
def move_exp_nansum(a, alpha, out):
    """
    Calculates the exponentially decayed sum.
    Very similar to move_exp_nanmean, but calculates a decayed sum rather than the mean.
    """

    # As per https://github.com/numbagg/numbagg/issues/26#issuecomment-830437132, we
    # could try implementing this with a bool flag on the same function as
    # `move_exp_nanmean` and hope that numba optimizes it away. The code here is
    # basically a subset of that function.

    N = len(a)
    if N == 0:
        return

    weight = 1.0 - alpha
    ignore_na = False  # could add as option in the future
    weighted_sum = 0

    for i in range(N):
        cur = a[i]
        is_observation = not np.isnan(cur)
        if not np.isnan(weighted_sum):
            if is_observation or (not ignore_na):
                weighted_sum = weight * weighted_sum
                if is_observation:
                    weighted_sum += cur
        elif is_observation:
            # The first non-nan value.
            weighted_sum = cur

        out[i] = weighted_sum
